/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

/* This module contains the thermodynamics for BEM. This is the
   interface between the chemical equilibrium model as defined in the
   model file and the numerical search algorithm. Function fxdfdx is
   called indirectly by the numerical method. Function ctrleq is used
   to contol the results. */

#include <stdio.h>
#include <math.h>

#include "bem.h"

#define TTI(x) printf(">>>>>>>>>>>>>>> %d\n",x);
#define TTX(x) printf("............... %g\n",x);




int fxdfdx(double xarg[MAXDIM], int nxarg, double fval[MAXDIM],
           double dfdx[MAXDIM][MAXDIM]){
/* This is a mapping of nxarg arguments (xarg) to nxarg values
   (fval). In case the argument vector represents an equilibrium, the
   values are all zero. The Jacobian is calculated and delivered in
   dfdx. */

/*
   chemical meaning of some functions and variables:
          
   i            elt        x                          f
---------------------------------------------------------------------    
 0               H         ln ((H+))   $           charge balance   $
 1               OH        ln (I)      $           I - balance      $
 2...nelt-1      others    ln ((ion of elt i)) $   elt-balance      #
 nelt-1+niex               ln (exch-H) /           CEC-balance      /
 nelt+niex...              ln (cslt)   %           ln(SI) or 0.0 
--------------------------------------------------------------------- 
$): mol/L      %): mol/kg    /): fraction   #): mol                   */

int i,ii,iii,m,z,k,kk,nn;
double a,b,bb,c,d,e,xoh;
double lngam[MAXCHRG+1],dlngam[MAXCHRG+1];

e=1;
for(i=0;i<nxarg;i++){
  fval[i]=0.;
  for(ii=0;ii<nxarg;ii++){
    dfdx[i][ii]=0.;
  }
}
/*....................Ionic Strength and Activities ..................*/
d=exp(xarg[1]);
c=sqrt(d);
a=-0.5*(c/(1.0+c) - 0.3*d)*M_LN10; /* Davies' equation used here */
b=-0.5*(1. / (2.*c*(1.+c)*(1.+c)) -0.3)*d*M_LN10; 
for(i=0;i<=MAXCHRG;i++){
  lngam[i]=a*i*i;
  dlngam[i]=b*i*i;
}
fval[1] = -d;
dfdx[1][1] = -d;
Cl->ionstr = d;

/*............................................... H+  ................*/
c=exp(xarg[0]-lngam[1]);
fval[0]+=c;
dfdx[0][0]+=c;
dfdx[0][1]-=c*dlngam[1];
fval[1]+=0.5*c;
dfdx[1][0]=0.5*c;
dfdx[1][1]-=0.5*c*dlngam[1];
Cl->el[0]->csng=c;
/*............................................... OH- ................*/
a = Cl->lnkw;
xoh=a-xarg[0];
c=exp(xoh-lngam[1]);
fval[0]-=c;
dfdx[0][0]+=c;
dfdx[0][1]+=c*dlngam[1];
fval[1]+=0.5*c;
dfdx[1][0]-=0.5*c;
dfdx[1][1]-=0.5*c*dlngam[1];
Cl->el[1]->csng=c;

/*......................................... other elements ............*/
for(i=2; i < Cl->nelt; i++){
  a=Cl->el[i]->ctoti;
  z=Cl->el[i]->chrg;
  m=abs(z);
  c=exp(xarg[i]-lngam[m]);
  d=c*Cl->theta;
  fval[0]+=z*c;
  dfdx[0][i]+=z*c;
  dfdx[0][1]-=z*c*dlngam[m];
  fval[1]+=0.5*m*m*c;
  dfdx[1][i]+=0.5*m*m*c;
  dfdx[1][1]-=0.5*m*m*c*dlngam[m];
  fval[i] = -a + d;
  dfdx[i][i] = d;
  dfdx[i][1] = -d*dlngam[m];
  Cl->el[i]->csng=c;
}

/*........................................... complexes .................*/
for(i=0;i < Cl->ncpl;i++){
  a = -Cl->cp[i]->lnkds;
  for(ii=0; ii < Cl->cp[i]->nelt; ii++){
    k = Cl->cp[i]->el[ii]->num;
    b = Cl->cp[i]->stcf[ii];
    if(k==1){
      a += b*xoh;
    }
    else{
      a += b*xarg[k];
    }
  }
  z=Cl->cp[i]->chrg;
  m=abs(z);
  c=exp(a - lngam[m]);

  d = c * Cl->theta ;
  if(Cl->cp[i]->gas != NULL){

    d += c * (Cl->sigma/Cl->cp[i]->gas->khen/RGAS/Cl->temp);
  }


  fval[0] += z*c;
  dfdx[0][1] += -z*c*dlngam[m];
  fval[1] += 0.5*m*m*c;
  dfdx[1][1] += -0.5*m*m*c*dlngam[m];
  for(ii=0;ii<Cl->cp[i]->nelt;ii++){
    k = Cl->cp[i]->el[ii]->num;
    b = Cl->cp[i]->stcf[ii];
    if(k==0){                 /* acid */
      dfdx[0][0] += z*c*b;
      dfdx[1][0] += 0.5*m*m*c*b;
    }
    if(k==1){                 /* OH - complex */
      dfdx[0][0] -= z*c*b;
      dfdx[1][0] -= 0.5*m*m*c*b;
    }
    if(k>1){
      dfdx[0][k] += z*c*b;
      dfdx[1][k] += 0.5*m*m*c*b;
      fval[k] += d*b; 
      dfdx[k][1] -= d*b*dlngam[m];
      for(iii=0;iii<Cl->cp[i]->nelt;iii++){
	kk = Cl->cp[i]->el[iii]->num;
	bb = Cl->cp[i]->stcf[iii];
	if(kk==1){
	  kk=0;
	  bb = -bb;
	}
	dfdx[k][kk] += d*b*bb;
      }
    }
  }
  Cl->cp[i]->ccpl = c;
}

/*......................................... cation exchange ..............*/
if(Cl->niex == 1){
  k=Cl->nelt;
  if(Cl->xc->xcap > 0.){
    fval[k]=-1.;
    for(i=0;i<Cl->nelt;i++){
      z = Cl->el[i]->chrg;
      if(z > 0){
	a = xarg[i] +z*xarg[k] -z*xarg[0] -z*Cl->xc->lnksel[i];
	c = exp(a);
	fval[k] += c;
	dfdx[k][i] += c;
	dfdx[k][k] += c*z;
	dfdx[k][0] += -c*z;
	if(i>1){
	  b = Cl->xc->xcap / ( (double) z) * Cl->rho;
	  fval[i] += c*b;
	  dfdx[i][i] += b*c;
	  dfdx[i][k] += b*c*z;
	  dfdx[i][0] += -b*c*z;
	}
	Cl->xc->xs[i] = c;
      }
    }

  }
  else{
    fval[k]=0.;
  }
}

/*........................................... salts     .................*/
for(i=0;i < Cl->nslt;i++){
  kk = Cl->nelt + Cl->niex + i;
  if(Cl->sl[i]->state == '+'){ /* .............. salt exists ................*/
    c=exp(xarg[kk]);
    spfun(i,c,&a,&b);
    fval[kk] = -a;     /* ln(SI) */
    dfdx[kk][kk] = -c*b; 
    for(ii=0; ii < Cl->sl[i]->nelt; ii++){
      k = Cl->sl[i]->el[ii]->num;
      b = Cl->sl[i]->stcf[ii];
      if(k==1){
	fval[kk] += b*xoh;
	dfdx[kk][0] += -b;
      }
      else{
	fval[kk] += b*xarg[k];
	dfdx[kk][k] += b;
      }
    }
    for(ii=0;ii<Cl->sl[i]->nelt;ii++){     /* element balances */
      k = Cl->sl[i]->el[ii]->num;
      b = Cl->sl[i]->stcf[ii];
      if(k>1){
	fval[k] += c*Cl->rho*b; 
	dfdx[k][kk] += c*Cl->rho*b;
      }
    }
    
    Cl->sl[i]->cslt = c;
  }
  else{ /*............................... salt currently not present .....*/
    fval[kk]=0.;
    Cl->sl[i]->cslt = 0.;
  }
}

/*........................................... partial pressures .........*/
for(i=0;i<Cl->ngas;i++){
  if(Cl->gs[i]->pfix > 0.){
    k=Cl->gs[i]->elnum;
    nn=Cl->gs[i]->cpnum;
    c=Cl->cp[nn]->ccpl;
    z=Cl->cp[nn]->chrg;
    m=abs(z);
    fval[k] = c - Cl->gs[i]->pfix * Cl->gs[i]->khen;
    for(ii=0;ii<nxarg;ii++){
      dfdx[k][ii]=0.;
    }
    dfdx[k][1] -= c*dlngam[m];
    for(ii=0;ii<Cl->cp[nn]->nelt;ii++){
      bb=Cl->cp[nn]->stcf[ii];
      kk=Cl->cp[nn]->el[ii]->num;
      if(kk==1){
	kk=0;
	bb = -bb;
      }
      dfdx[k][kk] += c*bb;
    }


  }
}

return 0;
}

/*------------------ spfun =  solubility product function ------------------*/
/* These functions describe the solubility product of a salt as
   functions of the concentration in the solid phase of the salt
   itself. par[0] should be the constant part of that function. */

int spfun( int k, double x, double *lnk, double *dlnk){
  int typ;
  double xx,a,b,c;

  typ=Cl->sl[k]->idspf;
  switch (typ){

  case 0:            /* sp = constant, par[0]=pKsp */
    *lnk = -Cl->sl[k]->spfpar[0]*M_LN10; 
    *dlnk = 0.;
    break;

  case 1:            /* pK = a + b * exp(c*x), */
                     /*      with x:= gain of salt since start [mmol/kg]*/
    xx= 1000*(x + Cl->sl[k]->coff - Cl->sl[k]->cslt0);
    a=Cl->sl[k]->spfpar[0];
    b=Cl->sl[k]->spfpar[1];
    c=Cl->sl[k]->spfpar[2];
    *lnk = -(a+b*exp(c*xx))*M_LN10;
    *dlnk = -M_LN10*b*exp(c*xx)*c*1000;
    break;
  case 2:            /* lin/spline/lin */
    xx= 1000*(x + Cl->sl[k]->coff - Cl->sl[k]->cslt0);
    getlspl(xx,&a,&b);
    *lnk=-M_LN10*a;
    *dlnk=-M_LN10*b*1000;
    break;
  }

  return 0;
}

/*----------------------------------------- ctrleq --------------------------*/
int ctrleq(){
  int i,ii,iii,z,pp,k,m;
  double r,sa,sc,s,f,a,b,c,d,lngam[MAXCHRG+1];
  double eps=1.e-15;

/*............................................. ionic charge balance.........*/
  sa = sc = 0.;
  for(i=0;i<Cl->nelt;i++){
    z=Cl->el[i]->chrg;
    if(z>0) sc += z*Cl->el[i]->csng;
    if(z<0) sa += z*Cl->el[i]->csng;
  }
  for(i=0;i<Cl->ncpl;i++){
    z=Cl->cp[i]->chrg;
    if(z>0) sc += z*Cl->cp[i]->ccpl;
    if(z<0) sa += z*Cl->cp[i]->ccpl;
  }
  if(fabs(sc)<eps) return 1000;
  r=fabs((sa+sc)/sc);
  if(r>Cl->relerr) return 1001;

/*............................................. ionic strength .........*/
  s = 0.;
  for(i=0;i<Cl->nelt;i++){
    z=Cl->el[i]->chrg;
    s += 0.5*z*z*Cl->el[i]->csng;
  }
  for(i=0;i<Cl->ncpl;i++){
    z=Cl->cp[i]->chrg;
    s += 0.5*z*z*Cl->cp[i]->ccpl;
  }
  f=Cl->ionstr;
  if(f<eps) return 2000;
  r=fabs((s-f)/f);
  if(r>Cl->relerr) return 2001;

/*.............................................. element balances .......*/
  for(i=2;i<Cl->nelt;i++){

    s=Cl->el[i]->csng * Cl->theta;  /*  single ion */

    pp=0; /* controlled by fixed partial pressure? */
    for(ii=0;ii<Cl->ngas;ii++){                          /* gas phase  */
      if(Cl->gs[ii]->elnum==i){
	if(Cl->gs[ii]->pfix>0.){
	  pp=1;
	  break;
	}
	else{
	  s += Cl->cp[Cl->gs[ii]->cpnum]->ccpl / Cl->gs[ii]->khen /RGAS/Cl->temp
               * Cl->sigma;
	}
      }
    }
    if(pp>0) continue;

    for(ii=0;ii<Cl->ncpl;ii++){                          /* complexes  */
      for(iii=0;iii<Cl->cp[ii]->nelt;iii++){
	if(Cl->cp[ii]->el[iii]->num == i){
	  s += Cl->cp[ii]->stcf[iii] * Cl->cp[ii]->ccpl * Cl->theta;
	}
      }
    }

    for(ii=0;ii<Cl->nslt;ii++){                            /*   salts */
      for(iii=0;iii<Cl->sl[ii]->nelt;iii++){
	if(Cl->sl[ii]->el[iii]->num == i){
	  s += Cl->sl[ii]->stcf[iii] * Cl->sl[ii]->cslt * Cl->rho;
	}
      }
    }

    z=Cl->el[i]->chrg;                     /* cation exchange complex */
    if(z>0 && Cl->niex>0){
      s += Cl->xc->xs[i] / z * Cl->xc->xcap *Cl->rho;
    }

    f=Cl->el[i]->ctoti;
    if(f<eps) return 3000;
    r=fabs((s-f)/f);
    if(r>Cl->relerr){
      return 3000+i;
    }
  }

/*...................................... cation exchange capacity .........*/
  if(Cl->niex>0){
    s=0.;
    for(i=0;i<Cl->nelt;i++){
      z=Cl->el[i]->chrg; 
      if(z>0) s += Cl->xc->xs[i];
    }

    f=1.;
    r=fabs((s-f)/f);
    if(r>Cl->relerr) return 4000;
  }


/*........................................ salt saturation indices ........*/
  d=Cl->ionstr;
  c=sqrt(d);
  a=-0.5*(c/(1.0+c) - 0.3*d)*M_LN10; /* Davies' equation used here */
  for(i=0;i<=MAXCHRG;i++){
    lngam[i]=a*i*i;
  }
  for(i=0;i < Cl->nslt;i++){
    if(Cl->sl[i]->state == '+'){
/*      c=Cl->sl[i]->cslt + Cl->sl[i]->coff; 17.12.95 */
      c=Cl->sl[i]->cslt ;
      spfun(i,c,&a,&b);
      s = -a;     /* ln(SI) */
      for(ii=0; ii < Cl->sl[i]->nelt; ii++){
	k = Cl->sl[i]->el[ii]->num;
	b = Cl->sl[i]->stcf[ii];
	m=abs(Cl->el[k]->chrg);
	s += b*(log(Cl->el[k]->csng) + lngam[m]);
      }
      
      if(s < -Cl->lsierr){
	return 8000+i; /*  undersaturation */
      }
      if( s >  Cl->lsierr){
	return 8100+i;          /* oversaturation */
      }
    }
  }

/*.......................................... fixed partial gas pressures ...*/
  for(i=0;i<Cl->ngas;i++){
    s=Cl->gs[i]->pfix;
    if(s>0){
      k=Cl->gs[i]->cpnum;
      f=Cl->cp[k]->ccpl / Cl->gs[i]->khen;
      r=fabs((s-f)/s);
      if(r>Cl->relerr) return 5000+i;
    }
  }


/*..........................................................................*/
  return 0;
}
